# This module is a convenience import to bring the gl namespace into another module
# usage: from pxng.opengl import *
from OpenGL.GL import GL_COLOR_BUFFER_BIT, GL_DEPTH_BUFFER_BIT
from OpenGL.GL import glTexEnvf, GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE
from OpenGL.GL import glClear, glClearColor, glColor
from OpenGL.GL import glLoadIdentity, glOrtho, glPushMatrix
from OpenGL.GL import glMatrixMode, GL_PROJECTION, GL_MODELVIEW, glPopMatrix
from OpenGL.GL import glViewport
from OpenGL.GL import glScalef, glTranslatef, glScale, glTranslate, glRotate, glRotatef
from OpenGL.GL import glBlendFunc, GL_BLEND, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA
from OpenGL.GL import GL_RGBA, GL_RGB, GL_ALPHA
from OpenGL.GL import GL_UNSIGNED_BYTE
from OpenGL.GL import GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_TEXTURE_MIN_FILTER
from OpenGL.GL import GL_TEXTURE_RECTANGLE
from OpenGL.GL import GL_NEAREST, GL_LINEAR
from OpenGL.GL import glEnable, glDisable
from OpenGL.GL import glGenTextures, glBindTexture, glTexImage2D
from OpenGL.GL import glTexSubImage2D, glTexParameterf, glTexParameteri
from OpenGL.GL import glBegin, glEnd, GL_QUADS, GL_LINES, GL_LINE, GL_LINE_STIPPLE
from OpenGL.GL import glLineStipple
from OpenGL.GL import glTexCoord2f, glTexCoord2i, glVertex
from OpenGL.GL import GL_COMPILE
from OpenGL.GL import glGenLists, glNewList, glEndList, glCallLists, glListBase



import freetype as ft
import numpy
from numpy.core.multiarray import ndarray


# This code is derived from the example in the *freetype-py* repository
# FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier


class Font:
    def __init__(self, filename, size):
        self._font_data: ndarray = None
        self.glyph_width = None
        self.glyph_height = None
        self._make_font(filename, size)

    def _make_font(self, filename, size):
        face = ft.Face(filename)
        face.set_char_size(size * 64)
        if not face.is_fixed_width:
            raise UserWarning('Font is not monotype')

        # Determine largest glyph size
        width, height, ascender, descender = 0, 0, 0, 0
        for c in range(32, 128):
            face.load_char(chr(c), ft.FT_LOAD_RENDER | ft.FT_LOAD_FORCE_AUTOHINT)
            bitmap = face.glyph.bitmap
            width = max(width, bitmap.width)
            ascender = max(ascender, face.glyph.bitmap_top)
            descender = max(descender, bitmap.rows - face.glyph.bitmap_top)
        height = ascender + descender

        # Generate texture data
        z = numpy.zeros((height * 6, width * 16), dtype=numpy.ubyte)
        for j in range(6):
            for i in range(16):
                pos = chr(32 + j * 16 + i)
                face.load_char(pos, ft.FT_LOAD_RENDER | ft.FT_LOAD_FORCE_AUTOHINT)
                bitmap = face.glyph.bitmap
                x = i * width + face.glyph.bitmap_left
                y = j * height + ascender - face.glyph.bitmap_top
                z[y:y + bitmap.rows, x:x + bitmap.width].flat = bitmap.buffer

        self._font_data = z
        self.glyph_width = width
        self.glyph_height = height



WHITE = (1, 1, 1)
LIGHT_GREY = (0.75, 0.75, 0.75)
GREY = (0.5, 0.5, 0.5)
DARK_GREY = (0.25, 0.25, 0.25)
BLACK = (0, 0, 0)

RED_PASTEL = (1, 0.75, 0.75)
LIGHT_RED = (1, 0.5, 0.5)
RED = (1, 0, 0)
DARK_RED = (0.5, 0, 0)
VERY_DARK_RED = (0.25, 0, 0)

GREEN_PASTEL = (0.75, 1, 0.75)
LIGHT_GREEN = (0.5, 1, 0.5)
GREEN = (0, 1, 0)
DARK_GREEN = (0, 0.5, 0)
VERY_DARK_GREEN = (0, 0.25, 0)

BLUE_PASTEL = (0.75, 0.75, 1)
LIGHT_BLUE = (0.5, 0.5, 1)
BLUE = (0, 0, 1)
DARK_BLUE = (0, 0, 0.5)
VERY_DARK_BLUE = (0, 0, 0.25)

YELLOW_PASTEL = (1, 1, 0.75)
LIGHT_YELLOW = (1, 1, 0.5)
YELLOW = (1, 1, 0)
DARK_YELLOW = (0.5, 0.5, 0)
VERY_DARK_YELLOW = (0.25, 0.25, 0)

MAGENTA_PASTEL = (1, 0.75, 1)
LIGHT_MAGENTA = (1, 0.5, 1)
MAGENTA = (1, 0, 1)
DARK_MAGENTA = (0.5, 0, 0.5)
VERY_DARK_MAGENTA = (0.25, 0, 0.25)

CYAN_PASTEL = (0.75, 1, 1)
LIGHT_CYAN = (0.5, 1, 1)
CYAN = (0, 1, 1)
DARK_CYAN = (0, 0.5, 0.5)
VERY_DARK_CYAN = (0, 0.25, 0.25)

ORANGE_PASTEL = (1, 0.825, 0.2)
LIGHT_ORANGE = (1, 0.75, 0.1)
ORANGE = (1, 0.5, 0)
DARK_ORANGE = (0.5, 0.25, 0)
VERY_DARK_ORANGE = (0.25, 0.125, 0)

PURPLE_PASTEL = (0.825, 0.2, 1)
LIGHT_PURPLE = (0.75, 0.1, 1)
PURPLE = (0.5, 0, 1)
DARK_PURPLE = (0.25, 0, 0.5)
VERY_DARK_PURPLE = (0.125, 0, 0.25)

AZURE_PASTEL = (0.2, 0.825, 1)
LIGHT_AZURE = (0.1, 0.75, 1.0)
AZURE = (0, 0.5, 1)
DARK_AZURE = (0, 0.25, 0.5)
VERY_DARK_AZURE = (0, 0.125, 0.25)

SPRING_GREEN_PASTEL = (0.2, 1, 0.85)
LIGHT_SPRING_GREEN = (0.1, 1, 0.75)
SPRING_GREEN = (0, 1, 0.5)
DARK_SPRING_GREEN = (0, 0.5, 0.25)
VERY_DARK_SPRING_GREEN = (0, 0.25, 0.125)

PALETTE = [
    [WHITE, LIGHT_GREY, GREY, DARK_GREY, BLACK],
    [RED_PASTEL, LIGHT_RED, RED, DARK_RED, VERY_DARK_RED],
    [MAGENTA_PASTEL, LIGHT_MAGENTA, MAGENTA, DARK_MAGENTA, VERY_DARK_MAGENTA],
    [PURPLE_PASTEL, LIGHT_PURPLE, PURPLE, DARK_PURPLE, VERY_DARK_PURPLE],
    [BLUE_PASTEL, LIGHT_BLUE, BLUE, DARK_BLUE, VERY_DARK_BLUE],
    [AZURE_PASTEL, LIGHT_AZURE, AZURE, DARK_AZURE, VERY_DARK_AZURE],
    [CYAN_PASTEL, LIGHT_CYAN, CYAN, DARK_CYAN, VERY_DARK_CYAN],
    [SPRING_GREEN_PASTEL, LIGHT_SPRING_GREEN, SPRING_GREEN, DARK_SPRING_GREEN, VERY_DARK_SPRING_GREEN],
    [GREEN_PASTEL, LIGHT_GREEN, GREEN, DARK_GREEN, VERY_DARK_GREEN],
    [YELLOW_PASTEL, LIGHT_YELLOW, YELLOW, DARK_YELLOW, VERY_DARK_YELLOW],
    [ORANGE_PASTEL, LIGHT_ORANGE, ORANGE, DARK_ORANGE, VERY_DARK_ORANGE],
]



import time
from pathlib import Path

import glfw
import pxng
import pxng.keys
from pxng.colors import WHITE
from pxng.opengl import *


class Window:

    def __init__(self, width, height, title='', **kwargs):
        """
        Creates a new window.

        After the successful creation of a window the loop has
        to be manually started with *start_event_loop*.
        To draw, use *set_update_handler* with a *fn(window)* callback.

        Parameters
        ----------
        width: int
            Width of window.
        height: int
            Height of window.
        title: str
            Title to show in title bar.
        **kwargs
            vsync: bool
                Turn off or on vsync. default=False
            resizable: bool
                Indicate if the windows should be resizable. default=False
            x_scale: float
                Scaling in 'x' direction. default=1
            y_scale: float
                Scaling in 'y' direction. default=1
            scale: float
                Sets 'x_scale' and 'y_scale' to the same value.
            hdpi: bool
                Use HDPI if available. Default=False
            color: tuple of float
                Set the background color. RGB or RGBA supported
        """
        if not glfw.init():
            raise UserWarning('Unable to initialize glfw')
        self.fps = 0
        self._handler = None
        self._title = title
        self.width = width
        self.height = height
        self.context = {}

        resizable = glfw.FALSE
        if 'resizable' in kwargs:
            resizable = glfw.TRUE if kwargs['resizable'] else glfw.FALSE
        glfw.window_hint(glfw.RESIZABLE, resizable)

        hdpi = glfw.FALSE
        if 'hdpi' in kwargs:
            hdpi = glfw.TRUE if kwargs['hdpi'] else glfw.FALSE
        glfw.window_hint(glfw.COCOA_RETINA_FRAMEBUFFER, hdpi)

        vsync = False
        if 'vsync' in kwargs:
            vsync = bool(kwargs['vsync'])

        self.x_scale = kwargs['x_scale'] if 'x_scale' in kwargs else 1
        self.y_scale = kwargs['y_scale'] if 'y_scale' in kwargs else 1
        if 'scale' in kwargs:
            self.x_scale = self.y_scale = kwargs['scale']

        self._window = glfw.create_window(width, height, title, None, None)

        self._color = (0, 0, 0, 1)
        if 'color' in kwargs:
            self._color = kwargs['color']

        if not self._window:
            glfw.terminate()
            raise UserWarning('Unable to create window')

        glfw.set_input_mode(self._window, glfw.STICKY_KEYS, glfw.TRUE)
        glfw.make_context_current(self._window)
        glfw.swap_interval(1 if vsync else 0)

        font_path = Path(__file__).parent / 'resources/fonts/C64_Pro_Mono-STYLE.ttf'
        # font_path = Path(__file__).parent / 'resources/fonts/JetBrainsMono-Bold.ttf'
        self._text_renderer = pxng.TextRenderer(pxng.Font(str(font_path), 8))
        self._elapsed_time = 0
        self._current_tint = WHITE
        self._key_poller = pxng.keys.KeyPoller()

    def start_event_loop(self):
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(0, self.width, self.height, 0, -1, 1)
        glViewport(0, 0, self.width, self.height)
        glMatrixMode(GL_MODELVIEW)
        self._loop()

    def key_state(self, key) -> pxng.keys.KeyState:
        return self._key_poller.key_state(key)

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, title):
        self._title = title
        glfw.set_window_title(self._window, self._title)

    @property
    def elapsed_time(self) -> float:
        return self._elapsed_time

    def _loop(self):
        frame_count = 0
        fps_now = time.time()
        elapsed_now = time.time()
        while not glfw.window_should_close(self._window):
            self._key_poller.poll_keys(self._window)
            glLoadIdentity()
            glScalef(self.x_scale, self.y_scale, 1)

            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
            glClearColor(*self._color)

            now = time.time()
            self._elapsed_time = (now - elapsed_now)
            elapsed_now = now
            # draw call
            if self._handler is not None:
                self._handler(self)

            # Swap front and back buffers
            glfw.swap_buffers(self._window)

            # Poll for and process events
            glfw.poll_events()
            frame_count += 1
            fps_time_delta = (time.time() - fps_now)
            self.fps = frame_count / fps_time_delta

            if fps_time_delta >= 1:
                title = f'{self.title} @ {self.fps:.0f} FPS'
                glfw.set_window_title(self._window, title)
                fps_now = time.time()
                frame_count = 0

        glfw.terminate()

    def close_window(self):
        glfw.set_window_should_close(self._window, glfw.TRUE)

    def set_update_handler(self, handler):
        self._handler = handler

    @property
    def tint(self):
        return self._current_tint

    @tint.setter
    def tint(self, color):
        self._current_tint = color

    def draw_sprite(self, x, y, sprite, scale=1.0, tint=None):
        self._set_color(tint)
        glPushMatrix()
        glTranslatef(x, y, 0)
        glScalef(scale, scale, 1)
        sprite.draw()
        glPopMatrix()

    def draw_text(self, x, y, text, scale=1.0, tint=None, angle=0):
        self._set_color(tint)
        self._text_renderer.draw_string(x, y, text, scale, angle)

    def draw_partial_sprite(self, x, y, sprite, sx, sy, sw, sh, scale=1.0, tint=None):
        self._set_color(tint)
        glPushMatrix()
        glTranslatef(x, y, 0)
        glScalef(scale, scale, 1)
        sprite.draw_partial(sx, sy, sw, sh)
        glPopMatrix()

    def draw_grid(self, size=10, tint=None, factor=4, pattern=0xAAAA):
        self._set_color(tint)
        glLineStipple(factor, pattern)
        glEnable(GL_LINE_STIPPLE)
        x = size
        y = size
        glBegin(GL_LINES)
        while x < self.width:
            glVertex(x, 0)
            glVertex(x, self.height)
            x += size
        while y < self.height:
            glVertex(0, y)
            glVertex(self.width, y)
            y += size
        glEnd()

    def fill_rect(self, x, y, w, h, tint=None):
        glPushMatrix()
        self._set_color(tint)
        glTranslatef(x, y, 0)
        glBegin(GL_QUADS)
        glVertex(0, h)
        glVertex(0, 0)
        glVertex(w, 0)
        glVertex(w, h)
        glEnd()
        glPopMatrix()

    def _set_color(self, tint):
        if tint is None:
            tint = self._current_tint
        glColor(*tint)



from . import opengl
from . import colors
from .font import Font
from .sprite import Sprite
from .animated_sprite import AnimatedSprite
from .display_list import DisplayList
from .text_renderer import TextRenderer
from .window import Window



from pxng.opengl import *


class DisplayList:
    def __init__(self, count=1):
        self._count = count
        self._created = False
        self._base = None
        self._next_list = 0
        self._create()

    def _create(self):
        self._base = glGenLists(self._count)
        self._created = True

    def start_display_list(self):
        #if not self._created:
        #    self._create()

        #if self._next_list >= self._count:
        #    raise UserWarning(f'max number of display lists reached: {self._count}')

        list_id = self._base + self._next_list
        glNewList(list_id, GL_COMPILE)
        self._next_list += 1
        return list_id

    def end_display_list(self):
        glEndList()

    def draw(self, indexes):
        glListBase(self._base)
        glCallLists(indexes)



from typing import Tuple

import imageio
from numpy.core.multiarray import ndarray
from pxng.opengl import *


class Sprite:
    def __init__(self, data: ndarray):
        self._data = data
        self._created = False
        self._dirty = False
        self._texid = None
        self._width = data.shape[1]
        self._height = data.shape[0]
        self._components = 1
        if len(data.shape) > 2:
            self._components = data.shape[2]

        if self._components == 4:
            self._format = GL_RGBA
        elif self._components == 3:
            self._format = GL_RGB
        elif self._components == 1:
            self._format = GL_ALPHA

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    def set_pixel(self, x, y, color: Tuple[int, int, int]):
        self._data[y, x] = color
        self._dirty = True

    def _create(self):
        self._texid = glGenTextures(1)
        glEnable(GL_TEXTURE_RECTANGLE)
        glBindTexture(GL_TEXTURE_RECTANGLE, self._texid)

        glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_NEAREST)

        d = self._data
        w = self._width
        h = self._height
        fmt = self._format
        glTexImage2D(GL_TEXTURE_RECTANGLE, 0, fmt, w, h, 0, fmt, GL_UNSIGNED_BYTE, d)
        self._created = True

    def _update(self):
        glEnable(GL_TEXTURE_RECTANGLE)
        glBindTexture(GL_TEXTURE_RECTANGLE, self._texid)

        d = self._data
        w = self._width
        h = self._height
        fmt = self._format
        glTexSubImage2D(GL_TEXTURE_RECTANGLE, 0, 0, 0, w, h, fmt, GL_UNSIGNED_BYTE, d)
        self._updated = True

    def update(self):
        self._dirty = True

    def activate(self):
        if not self._created:
            self._create()

        if self._dirty:
            self._update()

        glEnable(GL_TEXTURE_RECTANGLE)
        glBindTexture(GL_TEXTURE_RECTANGLE, self._texid)
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
        glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_NEAREST)

        if self._format in (GL_ALPHA, GL_RGBA):
            glEnable(GL_BLEND)
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    def deactivate(self):
        glDisable(GL_BLEND)
        glDisable(GL_TEXTURE_RECTANGLE)

    def _pre_draw(self):
        self.activate()
        glPushMatrix()

    def _post_draw(self):
        self.deactivate()
        glPopMatrix()

    def draw(self):
        """
        Draw the sprite at the current position.
        """
        self._pre_draw()
        glTranslate(0, self._height, 0)

        glBegin(GL_QUADS)
        glTexCoord2i(0, 0), glVertex(0, -self._height)
        glTexCoord2i(0, self._height), glVertex(0, 0)
        glTexCoord2i(self._width, self._height), glVertex(self._width, 0)
        glTexCoord2i(self._width, 0), glVertex(self._width, -self._height)
        glEnd()

        self._post_draw()

    def draw_partial(self, x, y, width, height):
        """
        Draw a partial sprite using pixel coordinates for the partial image data.
        The coordinates represent a sub region in the sprite.

        Parameters
        ----------
        x: int
            x coordinate in pixel space
        y: int
            y coordinate in pixel space
        width: int
            width in number of pixels
        height: int
            height in number of pixels
        """
        self._pre_draw()
        glTranslate(0, height, 0)  # anchor is upper left

        glBegin(GL_QUADS)
        glTexCoord2i(x, y), glVertex(0, -height)
        glTexCoord2i(x, y + height), glVertex(0, 0)
        glTexCoord2i(x + width, y + height), glVertex(width, 0)
        glTexCoord2i(x + width, y), glVertex(width, -height)
        glEnd()

        self._post_draw()

    @classmethod
    def create_from_image(cls, path):
        img_data = imageio.imread(path)
        return Sprite(img_data)



import pxng
from pxng.opengl import *


class TextRenderer:
    def __init__(self, font: pxng.Font, vertical_spacing=2):
        self._font = font
        self._vertical_spacing = vertical_spacing
        self._font_sprite = pxng.Sprite(self._font._font_data)
        self._display_list = pxng.DisplayList(8 * 16)
        self._initialized = False
        self._create_display_list()

    def _create_display_list(self):
        glPushMatrix()
        glLoadIdentity()
        sw = self._font.glyph_width
        sh = self._font.glyph_height
        for i in range(8 * 16):
            self._display_list.start_display_list()
            sx = i % 16
            sy = i // 16 - 2

            if i >= 32:
                glBegin(GL_QUADS)
                glTexCoord2i(sx * sw, sy * sh), glVertex(0, -sh)
                glTexCoord2i(sx * sw, sy * sh + sh), glVertex(0, 0)
                glTexCoord2i((sx + 1) * sw, sy * sh + sh), glVertex(sw, 0)
                glTexCoord2i((sx + 1) * sw, sy * sh), glVertex(sw, -sh)
                glEnd()

            glTranslatef(sw, 0, 0)
            self._display_list.end_display_list()
        glPopMatrix()
        self._initialized = True

    def draw_string(self, x, y, text, scale=1.0, angle=0):
        self._font_sprite.activate()

        glPushMatrix()
        sh = self._font.glyph_height
        glTranslatef(x, y, 0)
        glScalef(scale, scale, 1)
        glRotate(angle, 0, 0, 1)
        glTranslate(0, sh, 0)
        self._display_list.draw([ord(c) for c in text])
        glPopMatrix()
        self._font_sprite.deactivate()
